# IMPORTANT NOTES:
#     DO NOT MODIFY THIS FILE (please).
#
#     The functions in this file are an extract from scikit-image (v0.21.0), specifically:
#     skimage.exposure.equalize_adapthist
#     https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/exposure/_adapthist.py
#
#
# The purpose of this file is to expose: skimage.exposure.equalize_adaphist without
# all the dependencies of skimage.  E.g.
#
#     from teddy.utils.skimage_clahe import equalize_adapthist
#
# should work as a drop-in replacement for:
#
#     from skimage.exposure import equalize_adapthist
#


# LICENSE (copied from equalize_adapthist in scikit-image)
# License: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# .
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE HOLDERS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import math
import numbers
import warnings
from warnings import warn

import numpy as np

__all__ = ["equalize_adapthist"]

# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/util/dtype.py#L20
_integer_types = (
    np.byte,
    np.ubyte,  # 8 bits
    np.short,
    np.ushort,  # 16 bits
    np.intc,
    np.uintc,  # 16 or 32 or 64 bits
    int,
    np.int_,
    np.uint,  # 32 or 64 bits
    np.longlong,
    np.ulonglong,
)  # 64 bits

_integer_ranges = {
    t: (np.iinfo(t).min, np.iinfo(t).max) for t in _integer_types
}

dtype_range = {
    bool: (False, True),
    np.bool_: (False, True),
    float: (-1, 1),
    np.float_: (-1, 1),
    np.float16: (-1, 1),
    np.float32: (-1, 1),
    np.float64: (-1, 1),
}

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)

    if hasattr(np, "bool8"):
        dtype_range[np.bool8] = (False, True)

dtype_range.update(_integer_ranges)

_supported_types = list(dtype_range.keys())


NR_OF_GRAY = 2**14  # number of grayscale levels to use in CLAHE algorithm

DTYPE_RANGE = dtype_range.copy()
DTYPE_RANGE.update((d.__name__, limits) for d, limits in dtype_range.items())
DTYPE_RANGE.update(
    {
        "uint10": (0, 2**10 - 1),
        "uint12": (0, 2**12 - 1),
        "uint14": (0, 2**14 - 1),
        "bool": dtype_range[bool],
        "float": dtype_range[np.float64],
    }
)


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/_shared/utils.py#L533
def reshape_nd(arr, ndim, dim):
    if arr.ndim != 1:
        raise ValueError("arr must be a 1D array")
    new_shape = [1] * ndim
    new_shape[dim] = -1
    return np.reshape(arr, new_shape)


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/_shared/utils.py#L508
def slice_at_axis(sl, axis):
    return (slice(None),) * axis + (sl,) + (...,)


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/_shared/utils.py#L279
class channel_as_last_axis:
    def __init__(
        self,
        channel_arg_positions=(0,),
        channel_kwarg_names=(),
        multichannel_output=True,
    ):
        self.arg_positions = set(channel_arg_positions)
        self.kwarg_names = set(channel_kwarg_names)
        self.multichannel_output = multichannel_output

    def __call__(self, func):
        @functools.wraps(func)
        def fixed_func(*args, **kwargs):
            channel_axis = kwargs.get("channel_axis", None)

            if channel_axis is None:
                return func(*args, **kwargs)

            if np.isscalar(channel_axis):
                channel_axis = (channel_axis,)
            if len(channel_axis) > 1:
                raise ValueError(
                    "only a single channel axis is currently supported"
                )

            if channel_axis == (-1,) or channel_axis == -1:
                return func(*args, **kwargs)

            if self.arg_positions:
                new_args = []
                for pos, arg in enumerate(args):
                    if pos in self.arg_positions:
                        new_args.append(np.moveaxis(arg, channel_axis[0], -1))
                    else:
                        new_args.append(arg)
                new_args = tuple(new_args)
            else:
                new_args = args

            for name in self.kwarg_names:
                kwargs[name] = np.moveaxis(kwargs[name], channel_axis[0], -1)

            # now that we have moved the channels axis to the last position,
            # change the channel_axis argument to -1
            kwargs["channel_axis"] = -1

            # Call the function with the fixed arguments
            out = func(*new_args, **kwargs)
            if self.multichannel_output:
                out = np.moveaxis(out, -1, channel_axis[0])
            return out

        return fixed_func


def dtype_limits(image, clip_negative=False):
    imin, imax = dtype_range[image.dtype.type]
    if clip_negative:
        imin = 0
    return imin, imax


def _dtype_itemsize(itemsize, *dtypes):
    return next(dt for dt in dtypes if np.dtype(dt).itemsize >= itemsize)


def _dtype_bits(kind, bits, itemsize=1):
    s = next(
        i
        for i in (itemsize,) + (2, 4, 8)
        if bits < (i * 8) or (bits == (i * 8) and kind == "u")
    )

    return np.dtype(kind + str(s))


def _scale(a, n, m, copy=True):
    kind = a.dtype.kind
    if n > m and a.max() < 2**m:
        mnew = int(np.ceil(m / 2) * 2)
        if mnew > m:
            dtype = f"int{mnew}"
        else:
            dtype = f"uint{mnew}"
        n = int(np.ceil(n / 2) * 2)
        warn(
            f"Downcasting {a.dtype} to {dtype} without scaling because max "
            f"value {a.max()} fits in {dtype}",
            stacklevel=3,
        )
        return a.astype(_dtype_bits(kind, m))
    elif n == m:
        return a.copy() if copy else a
    elif n > m:
        # downscale with precision loss
        if copy:
            b = np.empty(a.shape, _dtype_bits(kind, m))
            np.floor_divide(
                a, 2 ** (n - m), out=b, dtype=a.dtype, casting="unsafe"
            )
            return b
        else:
            a //= 2 ** (n - m)
            return a
    elif m % n == 0:
        # exact upscale to a multiple of `n` bits
        if copy:
            b = np.empty(a.shape, _dtype_bits(kind, m))
            np.multiply(a, (2**m - 1) // (2**n - 1), out=b, dtype=b.dtype)
            return b
        else:
            a = a.astype(_dtype_bits(kind, m, a.dtype.itemsize), copy=False)
            a *= (2**m - 1) // (2**n - 1)
            return a
    else:
        # upscale to a multiple of `n` bits,
        # then downscale with precision loss
        o = (m // n + 1) * n
        if copy:
            b = np.empty(a.shape, _dtype_bits(kind, o))
            np.multiply(a, (2**o - 1) // (2**n - 1), out=b, dtype=b.dtype)
            b //= 2 ** (o - m)
            return b
        else:
            a = a.astype(_dtype_bits(kind, o, a.dtype.itemsize), copy=False)
            a *= (2**o - 1) // (2**n - 1)
            a //= 2 ** (o - m)
            return a


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/util/dtype.py#L188
def _convert(image, dtype, force_copy=False, uniform=False):
    image = np.asarray(image)
    dtypeobj_in = image.dtype
    if dtype is np.floating:
        dtypeobj_out = np.dtype("float64")
    else:
        dtypeobj_out = np.dtype(dtype)
    dtype_in = dtypeobj_in.type
    dtype_out = dtypeobj_out.type
    kind_in = dtypeobj_in.kind
    kind_out = dtypeobj_out.kind
    itemsize_in = dtypeobj_in.itemsize
    itemsize_out = dtypeobj_out.itemsize

    if np.issubdtype(dtype_in, np.obj2sctype(dtype)):
        if force_copy:
            image = image.copy()
        return image

    if not (dtype_in in _supported_types and dtype_out in _supported_types):
        raise ValueError(
            f"Cannot convert from {dtypeobj_in} to " f"{dtypeobj_out}."
        )

    if kind_in in "ui":
        imin_in = np.iinfo(dtype_in).min
        imax_in = np.iinfo(dtype_in).max
    if kind_out in "ui":
        imin_out = np.iinfo(dtype_out).min
        imax_out = np.iinfo(dtype_out).max

    # any -> binary
    if kind_out == "b":
        return image > dtype_in(dtype_range[dtype_in][1] / 2)

    # binary -> any
    if kind_in == "b":
        result = image.astype(dtype_out)
        if kind_out != "f":
            result *= dtype_out(dtype_range[dtype_out][1])
        return result

    # float -> any
    if kind_in == "f":
        if kind_out == "f":
            # float -> float
            return image.astype(dtype_out)

        if np.min(image) < -1.0 or np.max(image) > 1.0:
            raise ValueError("Images of type float must be between -1 and 1.")
        # floating point -> integer
        # use float type that can represent output integer type
        computation_type = _dtype_itemsize(
            itemsize_out, dtype_in, np.float32, np.float64
        )

        if not uniform:
            if kind_out == "u":
                image_out = np.multiply(
                    image, imax_out, dtype=computation_type
                )
            else:
                image_out = np.multiply(
                    image, (imax_out - imin_out) / 2, dtype=computation_type
                )
                image_out -= 1.0 / 2.0
            np.rint(image_out, out=image_out)
            np.clip(image_out, imin_out, imax_out, out=image_out)
        elif kind_out == "u":
            image_out = np.multiply(
                image, imax_out + 1, dtype=computation_type
            )
            np.clip(image_out, 0, imax_out, out=image_out)
        else:
            image_out = np.multiply(
                image,
                (imax_out - imin_out + 1.0) / 2.0,
                dtype=computation_type,
            )
            np.floor(image_out, out=image_out)
            np.clip(image_out, imin_out, imax_out, out=image_out)
        return image_out.astype(dtype_out)

    # signed/unsigned int -> float
    if kind_out == "f":
        # use float type that can exactly represent input integers
        computation_type = _dtype_itemsize(
            itemsize_in, dtype_out, np.float32, np.float64
        )

        if kind_in == "u":
            # using np.divide or np.multiply doesn't copy the data
            # until the computation time
            image = np.multiply(image, 1.0 / imax_in, dtype=computation_type)
            # DirectX uses this conversion also for signed ints
            # if imin_in:
            #     np.maximum(image, -1.0, out=image)
        elif kind_in == "i":
            # From DirectX conversions:
            # The most negative value maps to -1.0f
            # Every other value is converted to a float (call it c)
            # and then result = c * (1.0f / (2⁽ⁿ⁻¹⁾-1)).

            image = np.multiply(image, 1.0 / imax_in, dtype=computation_type)
            np.maximum(image, -1.0, out=image)

        else:
            image = np.add(image, 0.5, dtype=computation_type)
            image *= 2 / (imax_in - imin_in)

        return np.asarray(image, dtype_out)

    # unsigned int -> signed/unsigned int
    if kind_in == "u":
        if kind_out == "i":
            # unsigned int -> signed int
            image = _scale(image, 8 * itemsize_in, 8 * itemsize_out - 1)
            return image.view(dtype_out)
        else:
            # unsigned int -> unsigned int
            return _scale(image, 8 * itemsize_in, 8 * itemsize_out)

    # signed int -> unsigned int
    if kind_out == "u":
        image = _scale(image, 8 * itemsize_in - 1, 8 * itemsize_out)
        result = np.empty(image.shape, dtype_out)
        np.maximum(image, 0, out=result, dtype=image.dtype, casting="unsafe")
        return result

    # signed int -> signed int
    if itemsize_in > itemsize_out:
        return _scale(image, 8 * itemsize_in - 1, 8 * itemsize_out - 1)

    image = image.astype(_dtype_bits("i", itemsize_out * 8))
    image -= imin_in
    image = _scale(image, 8 * itemsize_in, 8 * itemsize_out, copy=False)
    image += imin_out
    return image.astype(dtype_out)


def _prepare_colorarray(arr, force_copy=False, *, channel_axis=-1):
    """Check the shape of the array and convert it to
    floating point representation.
    """
    arr = np.asanyarray(arr)

    if arr.shape[channel_axis] != 3:
        msg = (
            f"the input array must have size 3 along `channel_axis`, "
            f"got {arr.shape}"
        )
        raise ValueError(msg)

    float_dtype = _supported_float_type(arr.dtype)
    if float_dtype == np.float32:
        _func = img_as_float32
    else:
        _func = img_as_float64
    return _func(arr, force_copy=force_copy)


def _validate_channel_axis(channel_axis, ndim):
    if not isinstance(channel_axis, int):
        raise TypeError("channel_axis must be an integer")
    if channel_axis < -ndim or channel_axis >= ndim:
        raise np.AxisError("channel_axis exceeds array dimensions")


def img_as_float32(image, force_copy=False):
    return _convert(image, np.float32, force_copy)


def img_as_float64(image, force_copy=False):
    return _convert(image, np.float64, force_copy)


def img_as_float(image, force_copy=False):
    return _convert(image, np.floating, force_copy)


def img_as_uint(image, force_copy=False):
    return _convert(image, np.uint16, force_copy)


def img_as_int(image, force_copy=False):
    return _convert(image, np.int16, force_copy)


@channel_as_last_axis()
def rgb2hsv(rgb, *, channel_axis=-1):
    input_is_one_pixel = rgb.ndim == 1
    if input_is_one_pixel:
        rgb = rgb[np.newaxis, ...]

    arr = _prepare_colorarray(rgb, channel_axis=-1)
    out = np.empty_like(arr)

    # -- V channel
    out_v = arr.max(-1)

    # -- S channel
    delta = arr.ptp(-1)
    # Ignore warning for zero divided by zero
    old_settings = np.seterr(invalid="ignore")
    out_s = delta / out_v
    out_s[delta == 0.0] = 0.0

    # -- H channel
    # red is max
    idx = arr[..., 0] == out_v
    out[idx, 0] = (arr[idx, 1] - arr[idx, 2]) / delta[idx]

    # green is max
    idx = arr[..., 1] == out_v
    out[idx, 0] = 2.0 + (arr[idx, 2] - arr[idx, 0]) / delta[idx]

    # blue is max
    idx = arr[..., 2] == out_v
    out[idx, 0] = 4.0 + (arr[idx, 0] - arr[idx, 1]) / delta[idx]
    out_h = (out[..., 0] / 6.0) % 1.0
    out_h[delta == 0.0] = 0.0

    np.seterr(**old_settings)

    # -- output
    out[..., 0] = out_h
    out[..., 1] = out_s
    out[..., 2] = out_v

    # # remove NaN
    out[np.isnan(out)] = 0

    if input_is_one_pixel:
        out = np.squeeze(out, axis=0)

    return out


@channel_as_last_axis()
def hsv2rgb(hsv, *, channel_axis=-1):
    arr = _prepare_colorarray(hsv, channel_axis=-1)

    hi = np.floor(arr[..., 0] * 6)
    f = arr[..., 0] * 6 - hi
    p = arr[..., 2] * (1 - arr[..., 1])
    q = arr[..., 2] * (1 - f * arr[..., 1])
    t = arr[..., 2] * (1 - (1 - f) * arr[..., 1])
    v = arr[..., 2]

    hi = np.stack([hi, hi, hi], axis=-1).astype(np.uint8) % 6
    out = np.choose(
        hi,
        np.stack(
            [
                np.stack((v, t, p), axis=-1),
                np.stack((q, v, p), axis=-1),
                np.stack((p, v, t), axis=-1),
                np.stack((p, q, v), axis=-1),
                np.stack((t, p, v), axis=-1),
                np.stack((v, p, q), axis=-1),
            ]
        ),
    )
    return out


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/_shared/utils.py#L697
new_float_type = {
    # preserved types
    np.float32().dtype.char: np.float32,
    np.float64().dtype.char: np.float64,
    np.complex64().dtype.char: np.complex64,
    np.complex128().dtype.char: np.complex128,
    # altered types
    np.float16().dtype.char: np.float32,
    "g": np.float64,  # np.float128 ; doesn't exist on windows
    "G": np.complex128,  # np.complex256 ; doesn't exist on windows
}


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/_shared/utils.py#L710
def _supported_float_type(input_dtype, allow_complex=False):
    if isinstance(input_dtype, tuple):
        return np.result_type(*(_supported_float_type(d) for d in input_dtype))
    input_dtype = np.dtype(input_dtype)
    if not allow_complex and input_dtype.kind == "c":
        raise ValueError("complex valued input is not supported")
    return new_float_type.get(input_dtype.char, np.float64)


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/color/adapt_rgb.py#L12
def is_rgb_like(image, channel_axis=-1):
    """Return True if the image *looks* like it's RGB.

    This function should not be public because it is only intended to be used
    for functions that don't accept volumes as input, since checking an image's
    shape is fragile.
    """
    return (image.ndim == 3) and (image.shape[channel_axis] in (3, 4))


# From: https://github.com/scikit-image/scikit-image/blob/v0.21.0/skimage/color/adapt_rgb.py#L22
def adapt_rgb(apply_to_rgb):
    def decorator(image_filter):
        @functools.wraps(image_filter)
        def image_filter_adapted(image, *args, **kwargs):
            if is_rgb_like(image):
                return apply_to_rgb(image_filter, image, *args, **kwargs)
            else:
                return image_filter(image, *args, **kwargs)

        return image_filter_adapted

    return decorator


def hsv_value(image_filter, image, *args, **kwargs):
    # Slice the first three channels so that we remove any alpha channels.
    hsv = rgb2hsv(image[:, :, :3])
    value = hsv[:, :, 2].copy()
    value = image_filter(value, *args, **kwargs)
    hsv[:, :, 2] = _convert(value, hsv.dtype)
    return hsv2rgb(hsv)


def _output_dtype(dtype_or_range, image_dtype):
    if type(dtype_or_range) in [list, tuple, np.ndarray]:
        # pair of values: always return float.
        return _supported_float_type(image_dtype)
    if type(dtype_or_range) == type:
        # already a type: return it
        return dtype_or_range
    if dtype_or_range in DTYPE_RANGE:
        # string key in DTYPE_RANGE dictionary
        try:
            # if it's a canonical numpy dtype, convert
            return np.dtype(dtype_or_range).type
        except TypeError:  # uint10, uint12, uint14
            # otherwise, return uint16
            return np.uint16
    else:
        raise ValueError(
            "Incorrect value for out_range, should be a valid image data "
            f"type or a pair of values, got {dtype_or_range}."
        )


def intensity_range(image, range_values="image", clip_negative=False):
    if range_values == "dtype":
        range_values = image.dtype.type

    if range_values == "image":
        i_min = np.min(image)
        i_max = np.max(image)
    elif range_values in DTYPE_RANGE:
        i_min, i_max = DTYPE_RANGE[range_values]
        if clip_negative:
            i_min = 0
    else:
        i_min, i_max = range_values
    return i_min, i_max


def rescale_intensity(image, in_range="image", out_range="dtype"):
    if out_range in ["dtype", "image"]:
        out_dtype = _output_dtype(image.dtype.type, image.dtype)
    else:
        out_dtype = _output_dtype(out_range, image.dtype)

    imin, imax = map(float, intensity_range(image, in_range))
    omin, omax = map(
        float, intensity_range(image, out_range, clip_negative=(imin >= 0))
    )

    image = np.clip(image, imin, imax)

    if imin != imax:
        image = (image - imin) / (imax - imin)
        return (image * (omax - omin) + omin).astype(out_dtype)
    else:
        return np.clip(image, omin, omax).astype(out_dtype)


def _clahe(image, kernel_size, clip_limit, nbins):
    ndim = image.ndim
    dtype = image.dtype

    pad_start_per_dim = [k // 2 for k in kernel_size]

    pad_end_per_dim = [
        (k - s % k) % k + int(np.ceil(k / 2.0))
        for k, s in zip(kernel_size, image.shape)
    ]

    image = np.pad(
        image,
        [[p_i, p_f] for p_i, p_f in zip(pad_start_per_dim, pad_end_per_dim)],
        mode="reflect",
    )

    bin_size = 1 + NR_OF_GRAY // nbins
    lut = np.arange(NR_OF_GRAY, dtype=np.min_scalar_type(NR_OF_GRAY))
    lut //= bin_size

    image = lut[image]

    ns_hist = [int(s / k) - 1 for s, k in zip(image.shape, kernel_size)]
    hist_blocks_shape = np.array([ns_hist, kernel_size]).T.flatten()
    hist_blocks_axis_order = np.array(
        [np.arange(0, ndim * 2, 2), np.arange(1, ndim * 2, 2)]
    ).flatten()
    hist_slices = [
        slice(k // 2, k // 2 + n * k) for k, n in zip(kernel_size, ns_hist)
    ]
    hist_blocks = image[tuple(hist_slices)].reshape(hist_blocks_shape)
    hist_blocks = np.transpose(hist_blocks, axes=hist_blocks_axis_order)
    hist_block_assembled_shape = hist_blocks.shape
    hist_blocks = hist_blocks.reshape((math.prod(ns_hist), -1))

    # Calculate actual clip limit
    kernel_elements = math.prod(kernel_size)
    if clip_limit > 0.0:
        clim = int(np.clip(clip_limit * kernel_elements, 1, None))
    else:
        # largest possible value, i.e., do not clip (AHE)
        clim = kernel_elements

    hist = np.apply_along_axis(np.bincount, -1, hist_blocks, minlength=nbins)
    hist = np.apply_along_axis(clip_histogram, -1, hist, clip_limit=clim)
    hist = map_histogram(hist, 0, NR_OF_GRAY - 1, kernel_elements)
    hist = hist.reshape(hist_block_assembled_shape[:ndim] + (-1,))

    # duplicate leading mappings in each dim
    map_array = np.pad(
        hist, [[1, 1] for _ in range(ndim)] + [[0, 0]], mode="edge"
    )

    ns_proc = [int(s / k) for s, k in zip(image.shape, kernel_size)]
    blocks_shape = np.array([ns_proc, kernel_size]).T.flatten()
    blocks_axis_order = np.array(
        [np.arange(0, ndim * 2, 2), np.arange(1, ndim * 2, 2)]
    ).flatten()
    blocks = image.reshape(blocks_shape)
    blocks = np.transpose(blocks, axes=blocks_axis_order)
    blocks_flattened_shape = blocks.shape
    blocks = np.reshape(
        blocks, (math.prod(ns_proc), math.prod(blocks.shape[ndim:]))
    )

    # calculate interpolation coefficients
    coeffs = np.meshgrid(
        *tuple([np.arange(k) / k for k in kernel_size[::-1]]), indexing="ij"
    )
    coeffs = [np.transpose(c).flatten() for c in coeffs]
    inv_coeffs = [1 - c for dim, c in enumerate(coeffs)]

    result = np.zeros(blocks.shape, dtype=np.float32)
    for iedge, edge in enumerate(np.ndindex(*([2] * ndim))):
        edge_maps = map_array[
            tuple([slice(e, e + n) for e, n in zip(edge, ns_proc)])
        ]
        edge_maps = edge_maps.reshape((math.prod(ns_proc), -1))

        # apply map
        edge_mapped = np.take_along_axis(edge_maps, blocks, axis=-1)

        # interpolate
        edge_coeffs = np.prod(
            [[inv_coeffs, coeffs][e][d] for d, e in enumerate(edge[::-1])], 0
        )

        result += (edge_mapped * edge_coeffs).astype(result.dtype)

    result = result.astype(dtype)

    # rebuild result image from blocks
    result = result.reshape(blocks_flattened_shape)
    blocks_axis_rebuild_order = np.array(
        [np.arange(0, ndim), np.arange(ndim, ndim * 2)]
    ).T.flatten()
    result = np.transpose(result, axes=blocks_axis_rebuild_order)
    result = result.reshape(image.shape)

    # undo padding
    unpad_slices = tuple(
        [
            slice(p_i, s - p_f)
            for p_i, p_f, s in zip(
                pad_start_per_dim, pad_end_per_dim, image.shape
            )
        ]
    )
    result = result[unpad_slices]

    return result


@adapt_rgb(hsv_value)
def equalize_adapthist(image, kernel_size=None, clip_limit=0.01, nbins=256):
    float_dtype = _supported_float_type(image.dtype)
    image = img_as_uint(image)
    image = np.round(
        rescale_intensity(image, out_range=(0, NR_OF_GRAY - 1))
    ).astype(np.min_scalar_type(NR_OF_GRAY))

    if kernel_size is None:
        kernel_size = tuple([max(s // 8, 1) for s in image.shape])
    elif isinstance(kernel_size, numbers.Number):
        kernel_size = (kernel_size,) * image.ndim
    elif len(kernel_size) != image.ndim:
        raise ValueError(f"Incorrect value of `kernel_size`: {kernel_size}")

    kernel_size = [int(k) for k in kernel_size]

    image = _clahe(image, kernel_size, clip_limit, nbins)
    image = image.astype(float_dtype, copy=False)
    return rescale_intensity(image)


def clip_histogram(hist, clip_limit):
    # calculate total number of excess pixels
    excess_mask = hist > clip_limit
    excess = hist[excess_mask]
    n_excess = excess.sum() - excess.size * clip_limit
    hist[excess_mask] = clip_limit

    # Second part: clip histogram and redistribute excess pixels in each bin
    bin_incr = n_excess // hist.size  # average binincrement
    upper = clip_limit - bin_incr  # Bins larger than upper set to cliplimit

    low_mask = hist < upper
    n_excess -= hist[low_mask].size * bin_incr
    hist[low_mask] += bin_incr

    mid_mask = np.logical_and(hist >= upper, hist < clip_limit)
    mid = hist[mid_mask]
    n_excess += mid.sum() - mid.size * clip_limit
    hist[mid_mask] = clip_limit

    while n_excess > 0:  # Redistribute remaining excess
        prev_n_excess = n_excess
        for index in range(hist.size):
            under_mask = hist < clip_limit
            step_size = max(1, np.count_nonzero(under_mask) // n_excess)
            under_mask = under_mask[index::step_size]
            hist[index::step_size][under_mask] += 1
            n_excess -= np.count_nonzero(under_mask)
            if n_excess <= 0:
                break
        if prev_n_excess == n_excess:
            break

    return hist


def map_histogram(hist, min_val, max_val, n_pixels):
    out = np.cumsum(hist, axis=-1).astype(float)
    out *= (max_val - min_val) / n_pixels
    out += min_val
    np.clip(out, a_min=None, a_max=max_val, out=out)

    return out.astype(int)



import numpy as np
import pytest
import torch
import torchvision.transforms.functional as F
from PIL import Image

from chariot_transforms.preprocessing import (
    CLAHE,
    HistogramLinearStretch,
    IdentityPreprocessingTransform,
    PixelTransform,
    PreprocessingCompose,
    Resize,
    ResizePreserveAspect,
)
from chariot_transforms.preprocessing.clahe import equalize_adapthist
from chariot_transforms.preprocessing.histogram_tools import linear_stretch


@pytest.fixture
def img() -> torch.Tensor:
    return torch.randint(0, 256, (1, 3, 50, 150), dtype=torch.uint8) / 255


@pytest.fixture
def pil_img() -> Image.Image:
    rng = np.random.RandomState(2023)
    return Image.fromarray(
        rng.randint(0, 255, size=(500, 200, 3), dtype=np.uint8)
    )


@pytest.fixture
def mask() -> Image.Image:
    rng = np.random.RandomState(2023)
    return Image.fromarray(rng.randint(0, 255, size=(50, 150), dtype=np.uint8))


@pytest.fixture
def img_arr() -> np.ndarray:
    rng = np.random.RandomState(2023)
    return rng.randint(0, 256, (50, 50, 3), dtype=np.uint8)


def test_resize_regression(pil_img: Image.Image):
    """Test commutativity of resize and PIL -> tensor conversion which
    will ensure backwards compatibility for when we replace PIL transforms
    with tensor transforms. Some testing has shown that these operations are not
    close enough and downstream detections/classifications can vary depending
    on which one is used.
    """
    new_shape = (300, 600)
    img_tensor = F.to_tensor(pil_img).unsqueeze(0)

    resizer = Resize(new_shape)
    out_tensor = resizer(img_tensor)
    # this is what we had for PIL images before
    out_pil = resizer(pil_img)

    torch.testing.assert_close(
        out_tensor[0], F.to_tensor(out_pil), atol=0.004, rtol=0.004
    )


def test_resize_preserve_aspect_regression(pil_img: Image.Image):
    """Test commutativity of resize and PIL -> tensor conversion which
    will ensure backwards compatibility for when we replace PIL transforms
    with tensor transforms. Some testing has shown that these operations are not
    close enough and downstream detections/classifications can vary depending
    on which one is used.
    """

    new_shape = (300, 600)
    img_tensor = F.to_tensor(pil_img).unsqueeze(0)

    resizer = ResizePreserveAspect(*new_shape)
    out_tensor = resizer(img_tensor)
    out_pil = resizer(pil_img)

    torch.testing.assert_close(
        out_tensor[0], F.to_tensor(out_pil), atol=0.004, rtol=0.004
    )


def test_histogram_linear_stretch(img: torch.Tensor):
    ls = HistogramLinearStretch(4)
    out = ls(img)
    assert out.shape == img.shape
    assert out.min() >= 0
    assert out.max() <= 1
    assert not torch.equal(out, img)


def test_clahe(img: torch.Tensor):
    clahe = CLAHE()
    out = clahe(img)
    assert out.shape == img.shape
    assert out.min() >= 0
    assert out.max() <= 1
    assert not torch.equal(out, img)


def test_clahe_dtype(img_arr: np.ndarray):
    """Check that we get the same CLAHE normalized output for uint8 image arrays as
    float arrays
    """
    out1 = equalize_adapthist(img_arr)
    out2 = equalize_adapthist(img_arr.astype(np.float32) / 255.0)

    np.testing.assert_almost_equal(out1, out2, decimal=6)


def test_linear_stretch_dtype(img_arr: np.ndarray):
    """Check that we get the same linear stretch normalized output for uint8 image arrays as
    float arrays
    """
    out1 = linear_stretch(img_arr, 0.6)
    out2 = linear_stretch(img_arr.astype(np.float32) / 255.0, 0.6)

    np.testing.assert_almost_equal(out1, out2, decimal=6)


def test_resize_tensor_shape_error(img):
    resize = Resize((60, 30))
    with pytest.raises(ValueError) as exc_info:
        resize(img[0])

    assert "img must be a rank 4 tensor" in str(exc_info)


def test_resize_shapes(img: torch.Tensor):
    for s in [(60, 60), (200, 500)]:
        for cls in [Resize, ResizePreserveAspect]:
            resizer = cls(s) if cls == Resize else cls(*s)
            out_tensor = resizer(img)

            # Image.size is (width, height)
            assert out_tensor.shape[2:4] == s


def test_variable_shapes(img: torch.Tensor):
    for min_size, max_size in [
        (20, 30),
        (25, 75),
        (35, 200),
        (65, 100),
        (100, 250),
        (175, 300),
    ]:
        resizer = Resize(min_size=min_size, max_size=max_size)
        out_tensor = resizer(img)

        # Same checks on Tensor output
        assert (min(out_tensor.shape[2:4]) == min_size) or (
            max(out_tensor.shape[2:4]) == max_size
        ), "min_size: {0}, max_size: {1}".format(min_size, max_size)
        assert (
            max(out_tensor.shape[2:4]) <= max_size
        ), f"maximum output size too large, min_size: {min_size}, max_size: {max_size}"


def test_shapes_with_mask(img: torch.Tensor, mask: Image.Image):
    for s in [(60, 60), (200, 500)]:
        for cls in [Resize, ResizePreserveAspect]:
            resizer = cls(s) if cls == Resize else cls(*s)
            out_tensor, out_mask = resizer(img, mask=mask)

            # Confirm masks are resized correctly.
            assert out_tensor.shape[2:4] == (s[0], s[1])
            assert out_mask.size == (s[1], s[0])


def test_preprocessing_compose():
    assert PreprocessingCompose(
        [Resize((60, 60)), CLAHE()]
    ).resize_component == Resize((60, 60))

    assert PreprocessingCompose(
        [HistogramLinearStretch(0.5), ResizePreserveAspect(12, 20), CLAHE()]
    ).resize_component == ResizePreserveAspect(12, 20)

    assert PreprocessingCompose([CLAHE()]).resize_component is None

    with pytest.raises(ValueError) as exc_info:
        PreprocessingCompose(
            [Resize((60, 60)), CLAHE(), ResizePreserveAspect(12, 20)]
        )
    assert "cannot contain more than one resize transform" in str(exc_info)


def test___eq__():
    """Test equality of transforms"""
    r1 = Resize((60, 60))
    r2 = Resize((60, 60))
    r3 = Resize((60, 70))

    rpa1 = ResizePreserveAspect(60, 60)
    rpa2 = ResizePreserveAspect(60, 60)
    rpa3 = ResizePreserveAspect(10, 70)

    c1 = CLAHE()
    c2 = CLAHE(nbins=256)
    c3 = CLAHE(nbins=128)

    ls1 = HistogramLinearStretch(0.5)
    ls2 = HistogramLinearStretch(0.5)
    ls3 = HistogramLinearStretch(0.1)

    assert r1 == r2
    assert r1 != r3

    assert rpa1 == rpa2
    assert rpa1 != rpa3

    assert c1 == c2
    assert c1 != c3

    assert ls1 == ls2
    assert ls1 != ls3

    assert r1 != rpa1
    assert r1 != c1
    assert r1 != ls1
    assert rpa1 != c1
    assert rpa1 != ls1
    assert c1 != ls1


def test_multiple_channels():
    """Test that the pre-processing transformations work with
    more than 3 channel imagery
    """
    n_channels = 3
    img = torch.randint(0, 256, (n_channels, 50, 150), dtype=torch.uint8) / 255
    img = img.unsqueeze(0)

    # these transforms
    for transform, expected_shape in [
        (Resize((60, 60)), (60, 60)),
        (ResizePreserveAspect(new_h=40, new_w=60), (40, 60)),
        (HistogramLinearStretch(3), (50, 150)),
    ]:
        out = transform(img)
        assert out.shape == (1, n_channels, *expected_shape)

        # check that the resize transforms operate channel-wise
        if isinstance(transform, (Resize, ResizePreserveAspect)):
            for c in range(n_channels):
                torch.testing.assert_close(
                    transform(img[:, c : c + 1, :, :]), out[:, c : c + 1, :, :]
                )


def test_batching():
    imgs = torch.rand(2, 3, 50, 150)

    for transform in [
        Resize((60, 60)),
        ResizePreserveAspect(new_h=40, new_w=60),
        HistogramLinearStretch(3),
        CLAHE(),
        PixelTransform(mean=[0.5, 0.6, 0.2], std=[0.1, 1, 0.3]),
    ]:
        out = transform(imgs)
        # check that the resize transforms operate channel-wise
        for i in range(2):
            torch.testing.assert_close(
                transform(imgs[i : i + 1]), out[i : i + 1]
            )


def test_clahe_regression():
    """Regression test for CLAHE transform. guy_with_car_clahe.png was generated via
    ```
    from PIL import Image
    from chariot_transforms.preprocessing.transforms import IdentityPILTransform

    img = Image.open("tests/guy_with_car.jpg")
    clahe = IdentityPILTransform(histogram_equalization=True)
    clahe(img).save("tests/guy_with_car_clahe.png")
    ```
    """
    img = Image.open("tests/guy_with_car.jpg")
    img_tensor = F.to_tensor(img).unsqueeze(0)
    clahe = CLAHE()
    clahe_img_tensor = clahe(img_tensor)

    expected_img = Image.open("tests/guy_with_car_clahe.png")
    expected_img_tensor = F.to_tensor(expected_img).unsqueeze(0)

    assert (clahe_img_tensor - expected_img_tensor).abs().max().item() < 4e-3


def test_histogram_linear_stretch_regression():
    """Regression test for CLAHE transform. guy_with_car_clahe.png was generated via
    ```
    from PIL import Image
    from chariot_transforms.preprocessing.transforms import IdentityPILTransform

    img = Image.open("tests/guy_with_car.jpg")
    linear_stretch = IdentityPILTransform(histogram_linear_stretch=4)
    linear_stretch(img).save("tests/guy_with_car_linear_stretch.png")
    ```
    """
    img = Image.open("tests/guy_with_car.jpg")
    img_tensor = F.to_tensor(img).unsqueeze(0)
    linear_stretch = HistogramLinearStretch(4)
    linear_stretch_img_tensor = linear_stretch(img_tensor)

    expected_img = Image.open("tests/guy_with_car_linear_stretch.png")
    expected_img_tensor = F.to_tensor(expected_img).unsqueeze(0)

    assert (
        linear_stretch_img_tensor - expected_img_tensor
    ).abs().max().item() < 4e-3


def test_pixel_transform(img: torch.Tensor):
    mean = [0.5, 0.6, 0.2]
    std = [0.1, 1, 0.3]
    transform = PixelTransform(mean=mean, std=std)

    trans_img = transform(img)
    for c in range(img.shape[1]):
        for i in range(img.shape[2]):
            for j in range(img.shape[3]):
                assert (
                    trans_img[0, c, i, j]
                    == (img[0, c, i, j] - mean[c]) / std[c]
                )


def test_preprocessing_compose_equality():
    assert PreprocessingCompose(
        [
            IdentityPreprocessingTransform(),
            PixelTransform(mean=[0.5, 0.6, 0.2], std=[0.1, 1, 0.3]),
        ]
    ) == PixelTransform(mean=[0.5, 0.6, 0.2], std=[0.1, 1, 0.3])



import numpy as np


def linear_stretch(arr: np.ndarray, percentage: float) -> np.ndarray:
    """
    Applies a linear stretch to the histogram of arr (maintaining its shape).
    Returned array will have values in [0, 1].
    If percentage is close to 50 division by 0 may periodically occur.
    If percentage is larger than 50 returned values will be negative.
    It is recommended that percentage be in (0, 5].
    """
    lower, upper = np.percentile(arr.flatten(), [percentage, 100 - percentage])
    return np.clip((arr - lower) / (upper - lower), a_min=0, a_max=1)



"""
This script makes a grid for each augmentation in RandAugment. The first row
is with the lowest magnitude, the second is with a small magnitude, and the third
is with the highest magnitude. The resulting images will be in the `sample_imgs/augmentations` folder
"""


import os
from pathlib import Path

from PIL import Image
from torchvision.transforms.functional import to_pil_image, to_tensor
from torchvision.utils import make_grid

from chariot_transforms.augmentations.transforms import RandAugment

augmentations = RandAugment.augmentations

base_path = Path(__file__).parent
img = to_tensor(Image.open(base_path / "guy_with_car.jpg")).unsqueeze(0)

out_path = base_path / "augmentations"
os.makedirs(out_path, exist_ok=True)

bbox_dict = {
    "bboxes": [[64, 218, 250, 304], [87, 18, 271, 269]],
    "classes": ["person", "car"],
}


def get_percentile(high, low, p):
    return low + (high - low) * p


def draw_bounding_box_on_tensor(img, channel, ymin, xmin, ymax, xmax):
    img = img.clone()

    ymin = int(ymin)
    xmin = int(xmin)
    ymax = int(ymax)
    xmax = int(xmax)

    img[:, channel, ymin, xmin:xmax] = 1
    img[:, channel, ymax, xmin:xmax] = 1
    img[:, channel, ymin:ymax, xmin] = 1
    img[:, channel, ymin:ymax, xmax] = 1
    return img


def draw_bbox_dict_on_img(img, bbox_dict):
    class_to_channel = {"person": 0, "car": 1}
    for bbox, label in zip(bbox_dict["bboxes"], bbox_dict["classes"]):
        ymin, xmin, ymax, xmax = bbox
        img = draw_bounding_box_on_tensor(
            img, class_to_channel[label], ymin, xmin, ymax, xmax
        )
    return img


def apply_transform_and_draw_bbox(img, transform):
    trans_img, trans_bbox_dict = transform(img, bbox_dict=bbox_dict)
    trans_img_with_bboxes = draw_bbox_dict_on_img(trans_img, trans_bbox_dict)
    return trans_img_with_bboxes


n_iters = 10
for name, constr, (low, high) in augmentations:
    grid_imgs = []
    low_trans = constr(low)
    mid_trans = constr(get_percentile(high, low, 0.05))
    high_trans = constr(high)

    for t in [low_trans, mid_trans, high_trans]:
        for _ in range(n_iters):
            grid_imgs.append(apply_transform_and_draw_bbox(img, t)[0])

    grid = make_grid(grid_imgs, n_iters)
    to_pil_image(grid).save(out_path / f"{name}.png")



""" Joint versions of some of the transformations in torchvision.transforms
These are used, for example, in segmentation tasks where the input image and
target (mask) both need to be transformed in the same way.
"""
import math
import numbers
import random
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
from numpy import cos, sin, tan
from PIL import Image, ImageOps

try:
    import torchvision.transforms.v2.functional as F
    from torchvision.transforms.v2 import ColorJitter as _ColorJitter
    from torchvision.transforms.v2 import Grayscale as _Grayscale
except ModuleNotFoundError:
    import torchvision.transforms.functional as F
    from torchvision.transforms import (
        ColorJitter as _ColorJitter,
        Grayscale as _Grayscale,
    )

from chariot_transforms import bbox_ops
from chariot_transforms.base import (
    Compose,
    IdentityTransform,
    ImageOnlyTransform,
    TransformBase,
)
from chariot_transforms.preprocessing import (
    PreprocessingTransform,
    Resize,
    ResizePreserveAspect,
)

__all__ = [
    "AutoContrast",
    "Brightness",
    "Contrast",
    "Sharpness",
    "Invert",
    "Cutout",
    "Equalize",
    "Solarize",
    "Posterize",
    "Rotation",
    "RandomHorizontalFlip",
    "RandomVerticalFlip",
    "RandomCrop",
    "RandomRotation",
    "RandomPerspective",
    "RandomAffine",
    "ResizePreserveAspect",
    "RandomGrayscale",
    "ColorJitter",
    "Compose",
    "CenterCrop",
    "Resize",
    "RandomResize",
    "Grayscale",
    "PreprocessingTransform",
    "TranslateX",
    "TranslateY",
    "TransformBase",
    "Pad",
    "RandAugment",
]


def float_tensor_to_int_tensor(t: torch.Tensor) -> torch.Tensor:
    return (t * 255).to(torch.uint8)


def int_tensor_to_float_tensor(t: torch.Tensor) -> torch.Tensor:
    return t.to(torch.float32) / 255


def apply_along_classes(mask, fn):
    """
    Parameters
    ----------
    mask : Image.Image
    fn : callable

    Returns
    -------
    Image.Image
    """
    mask_array = np.array(mask)
    classes = [c for c in np.unique(mask_array) if c != 0]
    trans_mask = np.zeros((mask.height, mask.width), dtype=np.uint8)
    for c in classes:
        class_mask = np.zeros_like(mask_array, dtype=np.uint8)
        class_mask[mask_array == c] = 255

        trans_class_mask = np.array(fn(Image.fromarray(class_mask)))
        trans_class_mask[trans_class_mask < 255 // 2] = 0
        trans_class_mask[trans_class_mask >= 255 // 2] = c

        # pixels with different classes might go to the same pixel in
        # the resized image so we need to make sure we don't add these
        # classes together
        trans_mask[trans_class_mask != 0] = 0
        trans_mask += trans_class_mask

    return Image.fromarray(trans_mask)


def get_affine_matrix(
    center: Tuple[int, int],
    angle: float,
    translate: Tuple[int, int],
    scale: float,
    shear: Tuple[float, float],
):
    """
    Inverse of _get_inverse_affine_matrix from torchvision.transforms.functional

    Helper method to compute inverse matrix for affine transformation
    copied from torchvision.functional

    Computes transformation matrix: M = T * C * RSS * C^-1
    where T is translation matrix: [1, 0, tx | 0, 1, ty | 0, 0, 1]
          C is translation matrix to keep center: [1, 0, cx | 0, 1, cy | 0, 0, 1]
          RSS is rotation with scale and shear matrix
          RSS(a, s, (sx, sy)) =
          = R(a) * S(s) * SHy(sy) * SHx(sx)
          = [ s*cos(a - sy)/cos(sy), s*(-cos(a - sy)*tan(x)/cos(y) - sin(a)), 0 ]
            [ s*sin(a + sy)/cos(sy), s*(-sin(a - sy)*tan(x)/cos(y) + cos(a)), 0 ]
            [ 0                    , 0                                      , 1 ]

    where R is a rotation matrix, S is a scaling matrix, and SHx and SHy are the shears:
    SHx(s) = [1, -tan(s)] and SHy(s) = [1      , 0]
             [0, 1      ]              [-tan(s), 1]
    """

    if isinstance(shear, numbers.Number):
        shear = [shear, 0]

    if not isinstance(shear, (tuple, list)) and len(shear) == 2:
        raise ValueError(
            "Shear should be a single value or a tuple/list containing "
            + "two values. Got {}".format(shear)
        )

    rot = math.radians(angle)
    sx, sy = [math.radians(s) for s in shear]

    cx, cy = center
    tx, ty = translate

    # RSS without scaling
    a = cos(rot - sy) / cos(sy)
    b = -cos(rot - sy) * tan(sx) / cos(sy) - sin(rot)
    c = sin(rot - sy) / cos(sy)
    d = -sin(rot - sy) * tan(sx) / cos(sy) + cos(rot)

    # rotation matrix with scale and shear
    M = [a, b, 0, c, d, 0]
    M = [x * scale for x in M]

    # Apply inverse center translation
    M[2] += M[0] * (-cx) + M[1] * (-cy)
    M[5] += M[3] * (-cx) + M[4] * (-cy)

    # Apply translation and center translation
    M[2] += cx + tx
    M[5] += cy + ty
    M = M + [0, 0, 1]
    return np.array(M).reshape((3, -1))


class AugmentationTransform(TransformBase):
    # this class is just used for typing
    def __init__(self):
        pass


class IdentityAugmentationTransform(AugmentationTransform, IdentityTransform):
    def __init__(self):
        pass


class RandomTransform(AugmentationTransform):
    """Base class for transforms that have a random component to them."""

    def set_random_params(self, img):
        """Randomly sets the transformation parameters

        Parameters
        ----------
        img : PIL.Image
        """
        raise NotImplementedError

    def __call__(self, img, *args, **kwargs):
        self.set_random_params(img)
        return super().__call__(img=img, *args, **kwargs)


class AutoContrast(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = []

    def _convert_img_tensor(self, img):
        return F.autocontrast(img)

    def set_magnitude(self, magnitude: float) -> None:
        return


class Brightness(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        magnitude
            magnitude of the brightness, with 0.0 giving a black image and 1.0
            giving the original image.

            (For _convert_img_tensor) How much to adjust the brightness. Can be any non negative number.
            0 gives a black image, 1 gives the original image while 2 increases the brightness
            by a factor of 2
        """
        self.magnitude = magnitude

    def _convert_img_tensor(self, img: Image) -> Image:
        return F.adjust_brightness(img, brightness_factor=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class RandomBrightness(RandomTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        magnitude
            maximum magnitude of the brightness adjustment.

            Brightness adjusted by a factor between 1-magnitude and 1+magnitude.

            Required: magnitude must be >= 0.
        """
        if magnitude < 0:
            raise ValueError(
                "magnitude must be non-negative for brightness adjustment."
            )

        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return F.adjust_brightness(img, brightness_factor=self._magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude

    def set_random_params(self, img: Image) -> None:
        self._magnitude = np.random.uniform(
            low=max(0, 1 - self.magnitude), high=1 + self.magnitude
        )

    def _convert_bbox(self, bbox):
        return bbox

    def _convert_mask(self, mask):
        return mask


class Contrast(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        mangitude
            How much to adjust the contrast. Can be any non negative number.
            0 gives a solid gray image, 1 gives the original image while
            2 increases the contrast by a factor of 2
        """
        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor):
        return F.adjust_contrast(img, contrast_factor=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class RandomContrast(RandomTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        magnitude
            maximum magnitude of the contrast adjustment.

            Contrast adjusted by a factor between 1-magnitude and 1+magnitude.

            Required: magnitude must be >= 0.
        """
        if magnitude < 0:
            raise ValueError(
                "magnitude must be non-negative for contrast adjustment."
            )

        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor):
        return F.adjust_contrast(img, contrast_factor=self._magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude

    def set_random_params(self, img: Image) -> None:
        self._magnitude = np.random.uniform(
            low=max(0, 1 - self.magnitude), high=1 + self.magnitude
        )

    def _convert_bbox(self, bbox):
        return bbox

    def _convert_mask(self, mask):
        return mask


class Hue(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        mangitude
            How much to shift the hue channel. Should be in [-0.5, 0.5]. 0.5 and -0.5 give complete reversal
            of hue channel in HSV space in positive and negative direction respectively. 0 means no shift. Therefore,
            both -0.5 and 0.5 will give an image with complementary colors while 0 gives the original image.
        """
        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return F.adjust_hue(img, hue_factor=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class Gamma(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude", "gain"]

    def __init__(self, magnitude: float, gain: float = None) -> None:
        """
        Parameters
        ----------
        mangitude
            Non negative real number, same as γ in the equation. gamma larger than 1 make the shadows
            darker, while gamma smaller than 1 make dark regions lighter.
        gain
            The constant multiplier.
        """
        self.magnitude = magnitude
        self.gain = gain

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        if self.gain:
            return F.adjust_gamma(img, gamma=self.magnitude, gain=self.gain)
        else:
            return F.adjust_gamma(img, gamma=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class Saturation(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        mangitude
            How much to adjust the saturation. 0 will give a black and white image, 1 will give the
            original image while 2 will enhance the saturation by a factor of 2.
        """
        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return F.adjust_saturation(img, saturation_factor=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class Sharpness(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        mangitude
            magnitude of the sharpness, with 0.0 giving a blurred image, 1.0
            giving the original image, and 2.0 giving a sharpened image.
        """
        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor):
        return F.adjust_sharpness(img, sharpness_factor=self.magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude


class RandomSharpness(RandomTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """
        Parameters
        ----------
        magnitude
            maximum magnitude of the contrast adjustment.

            Contrast adjusted by a factor between 1-magnitude and 1+magnitude.

            Required: magnitude must be >= 0.
        """
        if magnitude < 0:
            raise ValueError(
                "magnitude for sharpness adjustment must be non-negative."
            )

        self.magnitude = magnitude

    def _convert_img_tensor(self, img: torch.Tensor):
        return F.adjust_sharpness(img, sharpness_factor=self._magnitude)

    def set_magnitude(self, magnitude: float) -> None:
        self.magnitude = magnitude

    def set_random_params(self, img: Image) -> None:
        self._magnitude = np.random.uniform(
            low=max(0, 1 - self.magnitude), high=1 + self.magnitude
        )

    def _convert_bbox(self, bbox):
        return bbox

    def _convert_mask(self, mask):
        return mask


class Equalize(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = []

    def _convert_img_tensor(self, img: torch.Tensor):
        return int_tensor_to_float_tensor(
            F.equalize(float_tensor_to_int_tensor(img))
        )

    def set_magnitude(self, magnitude: float) -> None:
        return


class Posterize(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["bits"]

    def __init__(self, bits: Union[int, float]) -> None:
        """
        Parameters
        ----------
        bits
            The number of bits to keep in each color channel.
            Should be between 1 and 8.
            Float values get converted to ints.

        Note: small values of bits have larger impact on the image than large values of bits
        because they result in more information being lost.
        """
        bits = int(bits)
        self.bits = bits

    def _convert_img_tensor(self, img):
        return int_tensor_to_float_tensor(
            F.posterize(float_tensor_to_int_tensor(img), self.bits)
        )


class RandomPosterize(RandomTransform):
    __serialization_attributes__ = ["bits"]

    def __init__(self, bits: Union[int, float]) -> None:
        """
        Parameters
        ----------
        bits
            The minimum number of bits to keep in each color channel.
            Should be between 1 and 8.
            Each random draw (for number of bits to keep) will be an integer
            between this number and 8.


        Note: small values of bits have larger impact on the image than large values of bits
        because they result in more information being lost.
        """
        self.bits = [i for i in range(1, 9) if i >= bits]
        if len(self.bits) == 0:
            raise ValueError(
                "The number of bits to keep should be between 1 and 8."
            )

    def set_random_params(self, img):
        self._bits = np.random.choice(self.bits)

    def convert_img(self, img):
        return ImageOps.posterize(img, self._bits)

    def _convert_img_tensor(self, img):
        return F.posterize(img, self._bits)

    def _convert_bbox(self, bbox):
        return bbox

    def _convert_mask(self, mask):
        return mask


class Solarize(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["threshold"]

    def __init__(self, threshold: Union[int, float]) -> None:
        """
        Parameters
        ----------
        threshold
            should be between 0 and 1. all values above this greyscale threshold
            are inverted.

        Note: smaller thresholds result in more pixels being inverted--a larger impact on the image.
        """
        self.threshold = threshold

    def _convert_img_tensor(self, img):
        # we need to add this because F.solarize with a threshold of 1 will send pixels with 1
        # to 0
        if self.threshold == 1:
            return img
        return F.solarize(img, self.threshold)


class RandomSolarize(RandomTransform):
    __serialization_attributes__ = ["threshold"]

    def __init__(self, threshold: Union[int, float]) -> None:
        """
        Parameters
        ----------
        threshold
            Minimum value for threshold for inversion.
            For each iteration, a random threshold will be chosen
            between this value and 256.
            All values above the this greyscale threshold are inverted.

            This value must be between 0 and 256.

        Note: smaller thresholds result in more pixels being inverted--a larger impact on the image.
        """
        if threshold < 0 or threshold > 1:
            raise ValueError("Solarizing threshold must be between 0 and 1.")
        self.threshold = threshold

    def set_random_params(self, img):
        self._threshold = np.random.uniform(low=self.threshold, high=1)

    def _convert_img_tensor(self, img):
        return F.solarize(img, self._threshold)

    def _convert_bbox(self, bbox):
        return bbox

    def _convert_mask(self, mask):
        return mask


class Invert(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = []

    def _convert_img_tensor(self, img):
        return F.invert(img)

    def set_magnitude(self, magnitude: float) -> None:
        return


class Cutout(RandomTransform):
    """Implementation of the data augmentation discussed in this paper:
    https://arxiv.org/abs/1708.04552
    """

    __serialization_attributes__ = ["length"]

    def __init__(self, length: int) -> None:
        """
        Parameters
        ----------
        length
            side length of the square to cutout. the original paper finds the
            optimal value via a grid search. their results are:

            - 16 for CIFAR10, 8 for CIFAR100
            - 20 for SVHN
            - 24 for STL10 if not using data augmentation, and 32 if training
              with augmentation.
        """
        self.length = length

    def set_random_params(self, img):
        h, w = img.shape[2:]
        center = np.array(
            [
                random.randint(0, h - 1),
                random.randint(0, w - 1),
            ]
        )
        self.ymin, self.xmin = np.clip(
            center - [self.length // 2, self.length // 2], a_min=0, a_max=None
        )
        self.ymax, self.xmax = np.clip(
            center + [self.length // 2, self.length // 2],
            a_min=None,
            a_max=(h, w),
        )

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        img = img.clone()

        if img.shape[0] != 1:
            raise ValueError(
                "Cutout currently only supports a batch of size 1."
            )

        img[0, :, self.ymin : self.ymax, self.xmin : self.xmax] = 0
        return img

    def _convert_mask(self, mask):
        return self.convert_img(mask)

    def _convert_bbox(self, bbox):
        cutout_bbox = [self.ymin, self.xmin, self.ymax, self.xmax]
        inter = bbox_ops.intersection(bbox, cutout_bbox)

        # If the bounding box does not overlap the cutout box, then just
        # return the bounding box.
        if inter == 0:
            return bbox

        area_bbox = bbox_ops.area(bbox)
        if inter == area_bbox:
            return None

        area_cutout = bbox_ops.area(cutout_bbox)
        if inter == area_cutout:
            return bbox

        if self.ymin <= bbox[0] and self.ymax >= bbox[2]:
            if self.xmin <= bbox[1]:
                return [bbox[0], self.xmax, bbox[2], bbox[3]]
            if self.xmax >= bbox[3]:
                return [bbox[0], bbox[1], bbox[2], self.xmin]

        if self.xmin <= bbox[1] and self.xmax >= bbox[3]:
            if self.ymin <= bbox[0]:
                return [self.ymax, bbox[1], bbox[2], bbox[3]]
            if self.ymax >= bbox[2]:
                return [bbox[0], bbox[1], self.ymin, bbox[3]]

        return bbox_ops.clean_bbox(bbox)


class Rotation(RandomTransform):
    __serialization_attributes__ = ["angle"]

    def __init__(self, angle: float, random_direction: bool = False) -> None:
        """
        Parameters
        ----------
        angle
            angle (in degrees) to rotate images by
        random_direction
            if True then will multiply angle by -1 with probability .5
        """
        self.angle = angle
        self.random_direction = random_direction

    def set_random_params(self, img):
        self._angle = self.angle
        if self.random_direction and random.random() < 0.5:
            self._angle = -self._angle

    def _convert_mask(self, mask):
        return self.convert_img(img=mask)

    def _convert_img_tensor(self, img):
        self._width = img.shape[-1]
        self._height = img.shape[-2]
        self._center = np.array([self._width / 2, self._height / 2])
        return F.rotate(img, self._angle)

    def _convert_bbox(self, bbox):
        ret = bbox_ops.rotate_bounding_box(self._angle, self._center, bbox)

        return bbox_ops.clean_bbox(ret, self._height, self._width, bbox)


class RandomVerticalFlip(RandomTransform):
    __serialization_attributes__ = ["p"]

    def __init__(self, p: float = 0.5):
        """
        Parameters
        ---------
        p : float
            probability of the image being flipped.
        self.p = p
        """
        self.p = p

    def set_random_params(self, img):
        self.flip = random.random() < self.p

    def _convert_mask(self, mask):
        return F.vflip(mask) if self.flip else mask

    def _convert_img_tensor(self, img):
        self.height = img.shape[2]
        return F.vflip(img) if self.flip else img

    def _convert_bbox(self, bbox):
        if self.flip:
            return [
                self.height - bbox[2],
                bbox[1],
                self.height - bbox[0],
                bbox[3],
            ]
        return bbox

    def __repr__(self):
        return self.__class__.__name__ + "(p={})".format(self.p)


class RandomHorizontalFlip(RandomTransform):
    """Horizontally flip the given PIL Images randomly with a given probability."""

    __serialization_attributes__ = ["p"]

    def __init__(self, p: float = 0.5):
        """
        Parameters
        ----------
        p : float
            probability of the image being flipped.
        """
        self.p = p

    def set_random_params(self, img):
        self.flip = random.random() < self.p

    def _convert_mask(self, mask):
        return F.hflip(mask) if self.flip else mask

    def _convert_img_tensor(self, img):
        # save the width since its needed for converting the bounding box
        self.width = img.shape[3]
        return F.hflip(img) if self.flip else img

    def _convert_bbox(self, bbox):
        if self.flip:
            return [
                bbox[0],
                self.width - bbox[3],
                bbox[2],
                self.width - bbox[1],
            ]
        return bbox

    def __repr__(self):
        return self.__class__.__name__ + "(p={})".format(self.p)


class RandomCrop(RandomTransform):
    """Crop the given PIL Images at a random location."""

    __serialization_attributes__ = ["size", "p"]

    def __init__(self, size, p: float = 1.0):
        """
        Parameters
        ----------
        size : sequence or int
            Desired output size of the crop. If size is an int instead of sequence
            like (h, w), a square crop (size, size) is made.
        p : float
          The probability of performing this operation at all
        """
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.p = p

    def set_random_params(self, img):
        h, w = img.shape[2:]
        self.th, self.tw = self.size

        # use relative sizes
        if isinstance(self.th, float):
            self.th = int(self.th * h)
        if isinstance(self.tw, float):
            self.tw = int(self.tw * w)

        if w == self.tw and h == self.th:
            self.i, self.j, self.th, self.tw = 0, 0, h, w
        if h - self.th >= 0 and w - self.tw >= 0:
            self.i = random.randint(0, h - self.th)
            self.j = random.randint(0, w - self.tw)
        else:
            self.tw = w
            self.th = h
            self.i = random.randint(0, h - self.th)
            self.j = random.randint(0, w - self.tw)

        self.do_transform = random.random() < self.p

    def _convert_img_tensor(self, img):
        if self.do_transform:
            return F.crop(img, self.i, self.j, self.th, self.tw)
        else:
            return img

    def _convert_mask(self, mask):
        if self.do_transform:
            return self.convert_img(img=mask)
        else:
            return mask

    def _convert_bbox(self, bbox):
        if self.do_transform:
            return bbox_ops.clean_bbox(
                [
                    bbox[0] - self.i,
                    bbox[1] - self.j,
                    bbox[2] - self.i,
                    bbox[3] - self.j,
                ],
                self.th,
                self.tw,
                bbox,
            )
        else:
            return bbox

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format(self.size)

    def __eq__(self, other):
        return (
            isinstance(other, RandomCrop)
            and self.size == other.size
            and self.p == other.p
        )


class RandomResize(RandomTransform):
    """Randomly rescale images jointly."""

    __serialization_attributes__ = [
        "scale_range_low",
        "scale_range_high",
        "p",
        "interpolation",
    ]

    def __init__(
        self,
        scale_range_low: float = 1,
        scale_range_high: float = 7,
        p: float = 0.5,
        interpolation: F.InterpolationMode = F.InterpolationMode.BILINEAR,
    ):
        """
        Parameters
        ----------
        scale_range_low : float
          The image will be resized to at a minimum of
          scale_range_low * L X scale_range_low * H
        scale_range_high : float
          The image will be resized to a maximum of
          scale_range_high * L X scale_range_high * H
        interpolation : PIL interpolation
          The kind of interpolation kernel to use.
        p : float
          The probability of doing a random resize.
        """
        self.scale_range_low = scale_range_low
        self.scale_range_high = scale_range_high
        self.interpolation = interpolation
        self.p = p

    def set_random_params(self, img):
        unif = random.uniform(0.0, 1.0)
        if unif < self.p:
            self.resize = random.uniform(
                self.scale_range_low, self.scale_range_high
            )
            self.size = (
                int(self.resize * img.shape[2]),
                int(self.resize * img.shape[3]),
            )
        else:
            self.resize = None
            self.size = None

    def _convert_img_tensor(self, img):
        if self.size:
            return F.resize(img, self.size, self.interpolation, antialias=True)
        return img

    def _convert_mask(self, mask):
        return self.convert_img(mask)

    def _convert_bbox(self, bbox):
        if self.size:
            return self.resize * np.array(bbox)
        return bbox

    def __repr__(self):
        return (
            self.__class__.__name__
            + "(low={0}, high={1}, interpolation={2}, p={3})".format(
                self.scale_range_low,
                self.scale_range_high,
                self.interpolation,
                self.p,
            )
        )


class RandomRotation(RandomTransform):
    """Rotate the images by a random angle."""

    __serialization_attributes__ = ["degrees"]

    def __init__(self, degrees: Union[float, List[float]]):
        """
        Parameters
        ----------
        degrees : sequence or float or int
            Range of degrees to select from. If degrees is a number instead of
            sequence like (min, max), the range of degrees will be (-degrees, +degrees).
        """
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError(
                    "If degrees is a single number, it must be positive."
                )
            self.degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError(
                    "If degrees is a sequence, it must be of len 2."
                )
            self.degrees = degrees

    def set_random_params(self, img):
        self.angle = np.random.uniform(
            low=self.degrees[0], high=self.degrees[1]
        )

    def _convert_mask(self, mask):
        return self.convert_img(img=mask)

    def _convert_img_tensor(self, img):
        self._width = img.shape[3]
        self._height = img.shape[2]
        self._center = np.array([self._width / 2, self._height / 2])
        return F.rotate(img, self.angle)

    def _convert_bbox(self, bbox):
        ret = bbox_ops.rotate_bounding_box(self.angle, self._center, bbox)

        return bbox_ops.clean_bbox(ret, self._height, self._width, bbox)

    def __repr__(self):
        format_string = self.__class__.__name__ + "(degrees={0}".format(
            self.degrees
        )
        format_string += ")"
        return format_string


class RandomPerspective(RandomTransform):
    """Performs Perspective transformation of the given PIL Images randomly
    with a given probability.
    """

    __serialization_attributes__ = ["distortion_scale", "p", "interpolation"]

    def __init__(
        self,
        distortion_scale: float = 0.5,
        p: float = 0.5,
        interpolation: F.InterpolationMode = F.InterpolationMode.BILINEAR,
    ):
        """
        Parameters
        ----------
        distortion_scale : float
            controls the degree of distortion and ranges from 0 to 1. Default value is 0.5.
        p : float
            probability of the image being perspectively transformed. Default value is 0.5
        interpolation : F.InterpolationMode
            interpolation method
        """
        self.p = p
        self.interpolation = interpolation
        self.distortion_scale = distortion_scale

    def set_random_params(self, img):
        self.apply_trans = True
        if random.random() >= self.p:
            self.apply_trans = False
        height, width = img.shape[2:]
        half_height = int(height / 2)
        half_width = int(width / 2)
        topleft = (
            random.randint(0, int(self.distortion_scale * half_width)),
            random.randint(0, int(self.distortion_scale * half_height)),
        )
        topright = (
            random.randint(
                width - int(self.distortion_scale * half_width) - 1, width - 1
            ),
            random.randint(0, int(self.distortion_scale * half_height)),
        )
        botright = (
            random.randint(
                width - int(self.distortion_scale * half_width) - 1, width - 1
            ),
            random.randint(
                height - int(self.distortion_scale * half_height) - 1,
                height - 1,
            ),
        )
        botleft = (
            random.randint(0, int(self.distortion_scale * half_width)),
            random.randint(
                height - int(self.distortion_scale * half_height) - 1,
                height - 1,
            ),
        )
        self.startpoints = [
            (0, 0),
            (width - 1, 0),
            (width - 1, height - 1),
            (0, height - 1),
        ]
        self.endpoints = [topleft, topright, botright, botleft]

    def _convert_img_tensor(self, img):
        if self.apply_trans:
            return F.perspective(
                img, self.startpoints, self.endpoints, self.interpolation
            )
        return img

    def _convert_mask(self, mask):
        if self.apply_trans:
            return apply_along_classes(
                mask,
                lambda x: F.perspective(
                    x, self.startpoints, self.endpoints, self.interpolation
                ),
            )
        return mask

    def _convert_bbox(self, bbox):
        # endpoints and startpoints arguments swapped since we need to get
        # the inverse of the transform
        a, b, c, d, e, f, g, h = self._get_perspective_coeffs(
            self.endpoints, self.startpoints
        )

        # transformation of a point (y, x)
        def trans(pt):
            # pt = [pt[1], pt[0]]
            s = g * pt[1] + h * pt[0] + 1
            return [
                (d * pt[1] + e * pt[0] + f) / s,
                (a * pt[1] + b * pt[0] + c) / s,
            ]

        # 4 corners
        points = [bbox[:2], bbox[2:], [bbox[0], bbox[3]], [bbox[2], bbox[1]]]
        new_points = np.array([trans(pt) for pt in points])
        minpts = new_points.min(0)
        maxpts = new_points.max(0)

        return np.array([minpts[0], minpts[1], maxpts[0], maxpts[1]])

    def __repr__(self):
        return self.__class__.__name__ + "(p={})".format(self.p)

    @staticmethod
    def _get_perspective_coeffs(startpoints, endpoints):
        """taken from torchvision.transforms.functional
        Helper function to get the coefficients (a, b, c, d, e, f, g, h) for the perspective transforms.

        In Perspective Transform each pixel (x, y) in the orignal image gets transformed as,
         (x, y) -> ( (ax + by + c) / (gx + hy + 1), (dx + ey + f) / (gx + hy + 1) )

        Parameters
        ----------
        startpoints : list
            List containing [top-left, top-right, bottom-right, bottom-left] of the orignal image,
        endpoints : list
            List containing [top-left, top-right, bottom-right, bottom-left] of the transformed
                       image
        Returns
        -------
            octuple (a, b, c, d, e, f, g, h) for transforming each pixel.
        """
        matrix = []

        for p1, p2 in zip(endpoints, startpoints):
            matrix.append(
                [p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]]
            )
            matrix.append(
                [0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]]
            )

        A = torch.tensor(matrix, dtype=torch.float)
        B = torch.tensor(startpoints, dtype=torch.float).view(8)
        # torch.lstq() is deprecated in 1.13.1
        # torch.lstq() solves minX|bx-a|
        # torch.linalg.lstsq() solves minX|ax-b|
        res = torch.linalg.lstsq(A, B)[0]
        return res.tolist()


class TranslateX(RandomTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """Translates an image horizontally (with the translation being to
        the left or right chosen randomly)

        Parameters
        ----------
        magnitude
            number between [0, 1] giving the maximum proportion of the image width to translate
        """
        if magnitude < 0 or magnitude > 1:
            raise ValueError("Translation proportions must be in [0, 1].")

        self.magnitude = magnitude

    def set_random_params(self, img):
        self._magnitude = np.random.uniform(
            low=-1 * self.magnitude, high=self.magnitude
        )

    def _convert_img_tensor(self, img):
        self._width = img.shape[3]
        self._translation = self._magnitude * self._width

        return F.affine(
            img, angle=0, translate=(self._translation, 0), scale=1, shear=0
        )

    def _convert_bbox(self, bbox):
        return bbox_ops.clean_bbox(
            [
                bbox[0],
                bbox[1] + self._translation,
                bbox[2],
                bbox[3] + self._translation,
            ],
            np.infty,
            self._width,
            bbox,
        )

    def _convert_mask(self, mask):
        return self.convert_img(mask)


class TranslateY(RandomTransform):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float) -> None:
        """Translates an image vertically (with the translation being to
        up or down chosen randomly)

        Parameters
        ----------
        magnitude
            number between [0, 1] giving the maximum proportion of the image height to translate
        """
        if magnitude < 0 or magnitude > 1:
            raise ValueError("Translation proportions must be in [0, 1].")

        self.magnitude = magnitude

    def set_random_params(self, img):
        self._magnitude = np.random.uniform(
            low=-1 * self.magnitude, high=self.magnitude
        )

    def _convert_img_tensor(self, img):
        self._height = img.shape[2]
        self._translation = self._magnitude * self._height

        return F.affine(
            img, angle=0, translate=(0, self._translation), scale=1, shear=0
        )

    def _convert_bbox(self, bbox):
        return bbox_ops.clean_bbox(
            [
                bbox[0] + self._translation,
                bbox[1],
                bbox[2] + self._translation,
                bbox[3],
            ],
            self._height,
            np.infty,
            bbox,
        )

    def _convert_mask(self, mask):
        return self.convert_img(mask)


class RandomAffine(RandomTransform):
    """Random affine transformation of the images keeping center invariant"""

    __serialization_attributes__ = [
        "degrees",
        "translate",
        "scale",
        "shear",
        "interpolation",
        "fill",
    ]

    def __init__(
        self,
        degrees: Union[float, List[float], Tuple[float, float]],
        translate: List[float] = None,
        scale: Tuple[float, float] = None,
        shear: Union[float, List[float]] = None,
        interpolation: F.InterpolationMode = F.InterpolationMode.NEAREST,
        fill: Union[int, tuple] = 0,
    ):
        """
        Parameters
        ----------
        degrees : sequence or float or int)
            Range of degrees to select from. If degrees is a number instead of
            sequence like (min, max), the range of degrees will be (-degrees, +degrees).
            Set to 0 to deactivate rotations.
        translate : tuple
            tuple of maximum absolute fraction for horizontal
            and vertical translations. For example translate=(a, b), then horizontal shift
            is randomly sampled in the range -img_width * a < dx < img_width * a and vertical shift is
            randomly sampled in the range -img_height * b < dy < img_height * b. Will not translate by default.
        scale : tuple
            scaling factor interval, e.g (a, b), then scale is
            randomly sampled from the range a <= scale <= b. Will keep original scale by default.
        shear : sequence or float or int
            Range of degrees to select from.
            If shear is a number, a shear parallel to the x axis in the range (-shear, +shear)
            will be apllied. Else if shear is a tuple or list of 2 values a shear parallel to the x axis in the
            range (shear[0], shear[1]) will be applied. Else if shear is a tuple or list of 4 values,
            a x-axis shear in (shear[0], shear[1]) and y-axis shear in (shear[2], shear[3]) will be applied.
            Will not apply shear by default
        interpolation (F.InterpolationMode): Desired interpolation enum defined by
            :class:`torchvision.transforms.F.InterpolationMode`. Default is ``F.InterpolationMode.NEAREST``.
            If input is Tensor, only ``F.InterpolationMode.NEAREST``, ``F.InterpolationMode.BILINEAR`` are supported.
            For backward compatibility integer values (e.g. ``PIL.Image[.Resampling].NEAREST``) are still accepted,
            but deprecated since Teddy 0.64.3 and will be removed soon! Please use F.InterpolationMode enum.
        fill (sequence or number): Pixel fill value for the area outside the transformed
            image. Default is ``0``. If given a number, the value is used for all bands respectively.
        """
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError(
                    "If degrees is a single number, it must be positive."
                )
            self.degrees = (-degrees, degrees)
        else:
            assert (
                isinstance(degrees, (tuple, list)) and len(degrees) == 2
            ), "degrees should be a list or tuple and it must be of length 2."
            self.degrees = degrees

        if translate is not None:
            assert (
                isinstance(translate, (tuple, list)) and len(translate) == 2
            ), "translate should be a list or tuple and it must be of length 2."
            for t in translate:
                if not (0.0 <= t <= 1.0):
                    raise ValueError(
                        "translation values should be between 0 and 1"
                    )
        self.translate = translate

        self.scale = scale
        if scale is not None:
            assert (
                isinstance(scale, (tuple, list)) and len(scale) == 2
            ), "scale should be a list or tuple and it must be of length 2."
            for s in scale:
                if s <= 0:
                    raise ValueError("scale values should be positive")
        self.scale_ranges = scale

        self.shear = shear
        if shear is not None:
            if isinstance(shear, numbers.Number):
                if shear < 0:
                    raise ValueError(
                        "If shear is a single number, it must be positive."
                    )
                self.shears = (-shear, shear)
            else:
                assert isinstance(shear, (tuple, list)) and (
                    len(shear) == 2 or len(shear) == 4
                ), "shear should be a list or tuple and it must be of length 2 or 4."
                # X-Axis shear with [min, max]
                if len(shear) == 2:
                    self.shears = [shear[0], shear[1], 0.0, 0.0]
                elif len(shear) == 4:
                    self.shears = [s for s in shear]
        else:
            self.shears = shear

        self.resample = False
        self.interpolation = interpolation
        self.fill = fill

    def set_random_params(self, img):
        height, width = img.shape[2:]
        angle = random.uniform(self.degrees[0], self.degrees[1])
        if self.translate is not None:
            max_dx = self.translate[0] * width
            max_dy = self.translate[1] * height
            translations = (
                np.round(random.uniform(-max_dx, max_dx)),
                np.round(random.uniform(-max_dy, max_dy)),
            )
        else:
            translations = (0, 0)

        if self.scale_ranges is not None:
            scale = random.uniform(self.scale_ranges[0], self.scale_ranges[1])
        else:
            scale = 1.0

        if self.shears is not None:
            if len(self.shears) == 2:
                shear = [random.uniform(self.shears[0], self.shears[1]), 0.0]
            elif len(self.shears) == 4:
                shear = [
                    random.uniform(self.shears[0], self.shears[1]),
                    random.uniform(self.shears[2], self.shears[3]),
                ]
        else:
            shear = 0.0

        self.angle = angle
        self.translations = translations
        self.scale = scale
        self.shear = shear

        center = (width * 0.5 + 0.5, height * 0.5 + 0.5)

        self.matrix = get_affine_matrix(
            center, angle, translations, scale, shear
        )

    def _convert_img_tensor(self, img):
        self._width = img.shape[3]
        self._height = img.shape[2]
        return F.affine(
            img,
            angle=self.angle,
            translate=self.translations,
            scale=self.scale,
            shear=self.shear,
            interpolation=(
                self.interpolation
                if not self.resample
                else F._interpolation_modes_from_int(self.resample)
            ),
            fill=self.fill,
        )

    def _convert_bbox(self, bbox):
        ret = bbox_ops.apply_affine_matrix_to_bbox(self.matrix, bbox)

        return bbox_ops.clean_bbox(ret, self._height, self._width, bbox)

    def _convert_mask(self, mask):
        return self.convert_img(mask)

    def __repr__(self):
        s = "{name}(degrees={degrees}"
        if self.translate is not None:
            s += ", translate={translate}"
        if self.scale_ranges is not None:
            s += ", scale={scale_ranges}"
        if self.shears is not None:
            s += ", shear={shears}"
        if self.interpolation:
            s += ", interpolation={interpolation}"
        if self.resample > 0:
            s += ", resample={resample}"
        if self.fillcolor:
            s += ", fillcolor={fillcolor}"
        if self.fill != 0:
            s += ", fill={fill}"
        s += ")"
        d = dict(self.__dict__)
        d["resample"] = str(d["resample"])
        d["interpolation"] = str(d["interpolation"])
        return s.format(name=self.__class__.__name__, **d)


class ColorJitter(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = [
        "brightness",
        "contrast",
        "saturation",
        "hue",
    ]

    def __init__(
        self,
        brightness: Union[float, Tuple[float, float]] = 0,
        contrast: Union[float, Tuple[float, float]] = 0,
        saturation: Union[float, Tuple[float, float]] = 0,
        hue: Union[float, Tuple[float, float]] = 0,
    ):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue
        self.transform = _ColorJitter(
            brightness=brightness,
            contrast=contrast,
            saturation=saturation,
            hue=hue,
        )

    def _convert_img_tensor(self, img):
        return self.transform(img)


class Grayscale(ImageOnlyTransform, AugmentationTransform):
    __serialization_attributes__ = ["num_output_channels"]

    def __init__(self, num_output_channels: int = 1):
        self.num_output_channels = num_output_channels
        self.transform = _Grayscale(num_output_channels=num_output_channels)

    def _convert_img_tensor(self, img):
        return self.transform(img)


class AugmentationCompose(Compose, RandomTransform):
    """Composes several transforms together."""

    def __init__(self, transforms: List[AugmentationTransform]):
        super().__init__(transforms)

    def set_random_params(self, img):
        for t in self.transforms:
            if isinstance(t, RandomTransform):
                t.set_random_params(img)


class CenterCrop(AugmentationTransform):
    """Crops the given PIL Images at the center."""

    __serialization_attributes__ = ["size"]

    def __init__(self, size: Union[int, List[int]]):
        """
        Parameters
        ----------
        size : sequence or int
            Desired output size of the crop. If size is an int instead of sequence
            like (h, w), a square crop (size, size) is made.
        """
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def _convert_mask(self, mask):
        return self.convert_img(mask)

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        # set the center_pt for the call to _convert_bbox
        h, w = img.shape[2:]
        self.center_pt = np.array([h / 2, w / 2])
        return F.center_crop(img, self.size)

    def _convert_bbox(self, bbox):
        pt1 = bbox[:2] - (self.center_pt - np.array(self.size) / 2)

        pt2 = bbox[2:] - (self.center_pt - np.array(self.size) / 2)

        ret = np.concatenate([pt1, pt2])
        return bbox_ops.clean_bbox(ret, self.size[0], self.size[1], bbox)

    def __repr__(self):
        return self.__class__.__name__ + "(size={0})".format(self.size)


class RandomGrayscale(RandomTransform):
    __serialization_attributes__ = ["p"]

    def __init__(self, p: float = 0.5):
        """
        Parameters
        ----------
        p : float
            The probability of converting to Grayscale
        """

        self.p = p

    def set_random_params(self, img):
        self.transform = IdentityTransform()
        if random.random() < self.p:
            self.transform = Grayscale(num_output_channels=3)

    def _convert_img_tensor(self, img):
        return self.transform(img)

    def _convert_mask(self, mask):
        return mask

    def _convert_bbox(self, bbox):
        return bbox

    def __repr__(self):
        return self.__class__.__name__ + "(p={})".format(self.p)


class _Shear(RandomTransform):
    def __init__(
        self, shear: Tuple[float, float], random_direction: bool = False
    ) -> None:
        """
        Parameters
        ----------
        shear : Tuple[float, float]
            Represents the maximum amount of shear to the image.
            The angle is non-zero in the first component for horizontal shear.
            The angle is non-zero in the second component for vertical shear.

            Note: Angles are randomly chosen up to the shear amount.
            If random_direction is True, then the angle is up to +- shear.

        random_direction : bool = False
            A boolean flag to indicate that randomly chosen shear amounts should
            be positive or negative.
        """
        self.shear = shear
        self.random_direction = random_direction

    def set_random_params(self, img):
        if self.random_direction:
            self._shear = (
                np.random.uniform(low=-1 * self.shear[0], high=self.shear[0]),
                np.random.uniform(low=-1 * self.shear[1], high=self.shear[1]),
            )
        else:
            self._shear = (
                np.random.uniform(low=0, high=self.shear[0]),
                np.random.uniform(low=0, high=self.shear[1]),
            )

    def _convert_bbox(self, bbox):
        ret = bbox_ops.apply_affine_matrix_to_bbox(self._matrix, bbox)
        return bbox_ops.clean_bbox(ret, self._height, self._width, bbox)

    def _convert_mask(self, mask):
        return self.convert_img(mask)

    def _convert_img_tensor(self, img):
        self._height, self._width = img.shape[2:]
        center = (self._width * 0.5 + 0.5, self._height * 0.5 + 0.5)
        self._matrix = get_affine_matrix(center, 0, (0, 0), 1.0, self._shear)
        return F.affine(
            img, angle=0, translate=(0, 0), scale=1, shear=self._shear
        )


class HorizontalShear(_Shear):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float):
        self.magnitude = magnitude
        super().__init__(
            (math.degrees(math.atan(magnitude)), 0), random_direction=True
        )


class VerticalShear(_Shear):
    __serialization_attributes__ = ["magnitude"]

    def __init__(self, magnitude: float):
        self.magnitude = magnitude
        super().__init__(
            (0, math.degrees(math.atan(magnitude))), random_direction=True
        )


class Pad(AugmentationTransform):
    """Wrapper for torchvision.transformations.Pad"""

    __serialization_attributes__ = ["padding", "fill", "padding_mode"]

    def __init__(
        self,
        padding: Union[
            int, List[int], Tuple[int, int], Tuple[int, int, int, int]
        ],
        fill: Union[int, Tuple[int, int, int]] = 0,
        padding_mode: str = "constant",
    ) -> None:
        """
        Parameters
        ----------
        padding
            Padding on each border. If a single int is provided this is used to pad all borders.
            If tuple of length 2 is provided this is the padding on left/right and top/bottom
            respectively. If a tuple of length 4 is provided this is the padding for the
            left, top, right and bottom borders respectively.
        fill
            Pixel fill value for constant fill. Default is 0. If a tuple of length 3, it is used
            to fill R, G, B channels respectively. This value is only used when the padding_mode
            is constant
        padding_mode
            Type of padding. Should be: constant, edge, reflect or symmetric.
            Default is constant. Mode symmetric is not yet supported for Tensor inputs.

            - constant: pads with a constant value, this value is specified with fill

            - edge: pads with the last value at the edge of the image

            - reflect: pads with reflection of image without repeating the last value on the edge

                For example, padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                will result in [3, 2, 1, 2, 3, 4, 3, 2]

            - symmetric: pads with reflection of image repeating the last value on the edge

                For example, padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                will result in [2, 1, 1, 2, 3, 4, 4, 3]
        """
        self.padding = padding
        self.fill = fill
        self.padding_mode = padding_mode

    def _convert_mask(self, mask):
        return self.convert_img(img=mask)

    def _convert_img_tensor(self, img):
        return F.pad(
            img,
            padding=self.padding,
            fill=self.fill,
            padding_mode=self.padding_mode,
        )

    def _convert_bbox(self, bbox):
        if self.padding_mode not in ["constant", "edge"]:
            raise RuntimeError(
                "Bounding box conversion for Pad transform is only supported when the "
                "padding mode is 'constant' or 'edge'."
            )

        if isinstance(self.padding, int):
            horizontal_shift = self.padding
            vertical_shift = self.padding
        else:  # case of tuple of length 2 or 4
            horizontal_shift, vertical_shift = self.padding[:2]

        return [
            bbox[0] + vertical_shift,
            bbox[1] + horizontal_shift,
            bbox[2] + vertical_shift,
            bbox[3] + horizontal_shift,
        ]


class RandAugment(RandomTransform):
    """Implementation of RandAugment (https://arxiv.org/abs/1909.13719)

    Attributes
    ----------
    augmentations : list[tuple]
        list of tuples of the form `(cls, (m_min, m_max))` where cls is the
        class of an augmentation, and m_min and m_max are lower and upper bounds,
        respectively, on the magnitude of the transformation. note that some
        transforms have no magnitude (e.g. AutoContrast, IdentityTransform)
        but bounds are included for uniformity.
    """

    # list of tuples of the form
    # (augmentation name, augmentation, (lower bound on magnitude, upper bound on magnitude)).
    augmentations = [
        ("AutoContrast", lambda x: AutoContrast(), (0.0, 1.0)),
        ("Brightness", RandomBrightness, (0.1, 0.9)),
        ("Contrast", RandomContrast, (0.1, 0.9)),
        ("Equalize", lambda x: Equalize(), (0.0, 1.0)),
        ("HorizontalShear", HorizontalShear, (0, 0.3)),
        ("Identity", lambda x: IdentityTransform(), (0.0, 1.0)),
        ("Invert", lambda x: Invert(), (0.0, 1.0)),
        ("Posterize", RandomPosterize, (8, 4)),
        ("Rotation", RandomRotation, (0.0, 30.0)),
        ("Sharpness", RandomSharpness, (0.1, 0.9)),
        ("Solarize", RandomSolarize, (1.0, 0.0)),
        ("TranslateX", TranslateX, (0.0, 1.0)),
        ("TranslateY", TranslateY, (0.0, 1.0)),
        ("VerticalShear", VerticalShear, (0.0, 0.3)),
    ]

    def __init__(
        self,
        n: int,
        m: Optional[int] = None,
        augs_to_use: Optional[List[str]] = None,
        augs_to_remove: Optional[List[str]] = None,
    ) -> None:
        """
        Parameters
        ----------
        n
            number of transforms to randomly draw
        m
            magnitude of the transforms. This should be between 0 and 10 if
            specified. If this is None then this is chosen randomly every
            time (this is used e.g. in FixMatch)
        augs_to_use
            list of augmentation names to use. If None then use all except for
            those in augs_to_remove
        augs_to_remove
            optional list of augmentations to not use.
        """
        self.n = n
        self.m = m
        self.augs_to_use = augs_to_use
        self.augs_to_remove = augs_to_remove

        def _validate_augmentation_names(aug_names: List[str]) -> None:
            # gives a warning if the user specifies the deprecated `Color` option and throws an error
            # if the user specifies an invalid augmentation name
            for aug_name in aug_names:
                if aug_name == "Color":
                    warnings.warn(
                        "The `Color` option has been removed from RandAugment so specifying it has no effect.",
                        DeprecationWarning,
                    )
                elif aug_name not in augmentation_names:
                    raise ValueError(
                        f"Invalid augmentation name '{aug_name}' in `augs_to_use`"
                    )

        augmentation_names = [a[0] for a in self.augmentations]
        if augs_to_use is not None:
            _validate_augmentation_names(augs_to_use)

            self.augmentations = [
                a for a in self.augmentations if a[0] in augs_to_use
            ]

        if augs_to_remove is not None:
            _validate_augmentation_names(augs_to_remove)
            self.augmentations = [
                a for a in self.augmentations if a[0] not in augs_to_remove
            ]

        if m is not None:
            if m > 10 or m < 0:
                raise ValueError("m must be between 0 and 10 (inclusive)")
            self.transforms = [
                t(b[0] + self.m / 10 * (b[1] - b[0]))
                for (_, t, b) in self.augmentations
            ]

    def set_random_params(self, img):
        """Sets self.transform to the random transform"""
        if self.m is not None:
            transforms = random.choices(self.transforms, k=self.n)
        else:
            transforms = [
                t(random.uniform(b[0], b[1]))
                for (_, t, b) in random.choices(self.augmentations, k=self.n)
            ]

        for t in transforms:
            if isinstance(t, RandomTransform):
                t.set_random_params(img)

        self._transform = AugmentationCompose(transforms)

    def _convert_img_tensor(self, *args, **kwargs):
        return self._transform._convert_img_tensor(*args, **kwargs)

    def _convert_bbox(self, bbox):
        return self._transform._convert_bbox(bbox)

    def _convert_mask(self, mask):
        return self._transform._convert_mask(mask)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RandAugment)
            and self.n == other.n
            and self.m == other.m
            and self.augs_to_use == other.augs_to_use
            and self.augs_to_remove == other.augs_to_remove
        )
